from fastapi.routing import APIRoute
from fastapi.openapi.models import APIKey
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from typing import List, Dict, Any
from pydantic import BaseModel
import logging
import traceback

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_request_body_example(route: APIRoute) -> Dict[str, Any]:
    """
    Get the example request body for a given route.

    Args:
        route (APIRoute): The route to get the request body example for.

    Returns:
        Dict[str, Any]: The example request body.
    """
    try:
        if route.body_field:
            if issubclass(route.body_field.type_, BaseModel):
                if route.body_field.field_info.examples:
                    example = next(
                        iter(route.body_field.field_info.examples.values())
                    ).get("value", {})
                else:
                    example = route.body_field.type_.model_json_schema().get(
                        "example", {}
                    )
                return example
            elif isinstance(route.body_field.type_, dict):
                return route.body_field.type_
            elif isinstance(route.body_field.type_, list):
                return [route.body_field.type_[0].model_json_schema()]
        return {}
    except Exception as e:
        logger.error(
            f"Error in get_request_body_example: {e}\n{traceback.format_exc()}"
        )
        return {}


def get_headers(route: APIRoute) -> List[Dict[str, str]]:
    """
    Get the headers required for a given route.

    Args:
        route (APIRoute): The route to get the headers for.

    Returns:
        List[Dict[str, str]]: The list of headers.
    """
    try:
        headers = []
        for dependency in route.dependant.dependencies:
            if isinstance(
                dependency.security_scopes,
                (
                    APIKey,
                    OAuth2PasswordBearer,
                    OAuth2PasswordRequestForm,
                ),
            ):
                headers.append(
                    {
                        "key": "Authorization",
                        "value": "Bearer {{access_token}}",
                    }
                )
        for param in route.dependant.header_params:
            headers.append(
                {
                    "key": param.name,
                    "value": f"{{{{{param.name}}}}}",
                }
            )
        return headers
    except Exception as e:
        logger.error(
            f"Error in get_headers: {e}\n{traceback.format_exc()}"
        )
        return []


def get_parameters(route: APIRoute) -> List[Dict[str, Any]]:
    """
    Get the parameters for a given route.

    Args:
        route (APIRoute):
        The route to get the parameters for.

    Returns:
        List[Dict[str, Any]]:
        The list of parameters.
    """
    try:
        parameters = []
        for param in route.dependant.query_params:
            parameters.append(
                {
                    "name": param.name,
                    "in": "query",
                    "required": param.required,
                    "schema": {
                        "type": (
                            param.type_.__name__
                            if hasattr(
                                param.type_,
                                "__name__"
                            )
                            else str(param.type_)
                        ),
                        "description": (
                            param.field_info.description
                            if param.field_info.description
                            else ""
                        ),
                        "default": (
                            param.default
                            if param.default is not None
                            else ""
                        ),
                        "example": (
                            param.field_info.extra.get(
                                "example", ""
                            )
                            if hasattr(
                                param.field_info,
                                "extra"
                            )
                            else ""
                        ),
                    },
                }
            )
        for param in route.dependant.path_params:
            parameters.append(
                {
                    "name": param.name,
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": (
                            param.type_.__name__
                            if hasattr(
                                param.type_,
                                "__name__"
                            )
                            else str(param.type_)
                        ),
                        "description": (
                            param.field_info.description
                            if param.field_info.description
                            else ""
                        ),
                        "default": (
                            param.default
                            if param.default is not None
                            else ""
                        ),
                        "example": (
                            param.field_info.extra.get(
                                "example",
                                ""
                            )
                            if hasattr(
                                param.field_info,
                                "extra"
                            )
                            else ""
                        ),
                    },
                }
            )
        return parameters
    except Exception as e:
        logger.error(
            f"Error in get_parameters: {e}\n{traceback.format_exc()}"
        )
        return []


def get_responses(route: APIRoute) -> Dict[str, Dict[str, Any]]:
    """
    Get the responses for a given route.

    Args:
        route (APIRoute): The route to get the responses for.

    Returns:
        Dict[str, Dict[str, Any]]: The dictionary of responses.
    """
    try:
        responses = {}
        if route.responses:
            for status_code, response in route.responses.items():
                if isinstance(response, dict):
                    content = (
                        response.get("content", {})
                        .get("application/json", {})
                        .get("schema", {})
                    )
                    if not content and "model" in response:
                        content = response["model"].model_json_schema()
                    responses[str(status_code)] = {
                        "description": response.get(
                            "description",
                            ""
                        ),
                        "content": {
                            "application/json": {
                                "schema": content
                            }
                        },
                    }
                elif isinstance(response, BaseModel):
                    responses[str(status_code)] = {
                        "description": response.__doc__,
                        "content": {
                            "application/json": {
                                "schema": response.model_json_schema()
                            }
                        },
                    }
        elif route.response_model:
            if isinstance(route.response_model, list):
                schema = [
                    model.model_json_schema()
                    for model in route.response_model
                ]
            else:
                schema = route.response_model.model_json_schema()
            responses[str(route.status_code)] = {
                "description": route.response_model.__doc__,
                "content": {
                    "application/json": {
                        "schema": schema
                    }
                },
            }
        return responses
    except Exception as e:
        logger.error(
            f"Error in get_responses: {e}\n{traceback.format_exc()}"
        )
        return {}



import json
import logging
from fastapi import FastAPI
from fastapi.routing import APIRoute
from .utils import (
    get_request_body_example,
    get_headers,
    get_parameters,
    get_responses,
)
from typing import Any, Dict
from fastapi.encoders import jsonable_encoder

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def generate_postman_collection(
    app: FastAPI,
    output_file: str = "postman_collection.json",
    input_name: str = "API Collection",
    input_host: str = "http://localhost",
    readme_file: str = "README.md",
) -> None:
    """
    Generate a Postman collection from a FastAPI app.

    Args:
        app (FastAPI):
            The FastAPI app instance.
        output_file (str):
            The output file name for the Postman collection.
        input_name (str):
            The name of the Postman collection.
        input_host (str):
            The host URL for the API.
        readme_file (str):
            The path to the README.md file for documentation.
    """
    try:
        with open(readme_file, "r") as f:
            readme_content = f.read()
    except Exception as e:
        logger.error(
            f"Error reading README.md file: {e}"
        )
        readme_content = ""

    collection: Dict[str, Any] = {
        "info": {
            "name": input_name,
            "schema": (
                "https://schema.getpostman.com/json/collection/v2.1.0/"
                "collection.json"
            ),
            "description": readme_content,
        },
        "item": [],
        "auth": {
            "type": "bearer",
            "bearer": [
                {
                    "key": "token",
                    "value": "{{access_token}}",
                    "type": "string",
                }
            ],
        },
    }

    folders = {}

    for route in app.routes:
        if isinstance(route, APIRoute):
            try:
                item = {
                    "name": route.name,
                    "request": {
                        "url": f"{input_host}{route.path}",
                        "method": route.methods.pop(),
                        "description": route.summary or "",
                        "header": get_headers(
                            route
                        ),
                        "body": {
                            "mode": "raw",
                            "raw": jsonable_encoder(
                                get_request_body_example(
                                    route
                                )
                            ),
                        },
                        "params": jsonable_encoder(
                            get_parameters(
                                route
                            )
                        ),
                        "responses": jsonable_encoder(
                            get_responses(
                                route
                            )
                        ),
                    },
                }
                for tag in route.tags:
                    if tag not in folders:
                        folders[tag] = {"name": tag, "item": []}
                    folders[tag]["item"].append(item)
            except Exception as e:
                logger.error(
                    f"Error processing route {route}: {e}"
                )

    collection["item"] = list(folders.values())

    try:
        with open(output_file, "w") as f:
            json.dump(collection, f, indent=4)
        logger.info(
            f"Postman collection saved to {output_file}"
        )
    except Exception as e:
        logger.error(
            f"Error saving Postman collection to {output_file}: {e}"
        )


def main() -> None:
    """
    Main function to parse arguments
    and generate the Postman collection.
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Generate Postman collection from FastAPI app."
    )
    parser.add_argument(
        "--app",
        required=True,
        help="Path to the FastAPI app",
    )
    parser.add_argument(
        "--output",
        default="postman_collection.json",
        help="Output file for the Postman collection",
    )
    parser.add_argument(
        "--name",
        default="API Collection",
        help="Name of the Postman collection",
    )
    parser.add_argument(
        "--host",
        default="http://localhost",
        help="Host URL for the API",
    )
    parser.add_argument(
        "--readme",
        default="README.md",
        help="Path to the README.md file for documentation",
    )

    args = parser.parse_args()

    try:
        app_module, app_var = args.app.split(":")
        app = getattr(
            __import__(
                app_module,
                fromlist=[app_var]
            ),
            app_var
        )

        generate_postman_collection(
            app, args.output, args.name, args.host, args.readme
        )
    except Exception as e:
        logger.error(
            f"Error importing FastAPI app from {args.app}: {e}"
        )


if __name__ == "__main__":
    main()



import json
from fastapi import (
    Body,
    Cookie,
    Depends,
    FastAPI,
    Header,
    HTTPException,
    Path,
    Query,
    Security,
    status,
)
from fastapi.security import (
    APIKeyHeader,
    OAuth2PasswordBearer,
    OAuth2PasswordRequestForm,
)
from fastapi.testclient import TestClient
from fast_man.converter import generate_postman_collection
from pydantic import BaseModel, Field
import pytest
from typing import Optional

# Security schemes
oauth2_scheme = OAuth2PasswordBearer(
    tokenUrl="token"
)
api_key_scheme = APIKeyHeader(
    name="X-API-Key"
)


# Models
class Item(BaseModel):
    name: str = Field(
        ...,
        json_schema_extra="Item name"
    )
    description: Optional[str] = Field(
        None,
        json_schema_extra="Item description"
    )


class ResponseItem(BaseModel):
    name: str = Field(
        ...,
        json_schema_extra="Item name"
    )
    description: Optional[str] = Field(
        None,
        json_schema_extra="Item description"
    )
    id: int = Field(
        ...,
        json_schema_extra=1
    )


class ErrorResponse(BaseModel):
    detail: str = Field(
        ...,
        json_schema_extra="Error detail"
    )


class User(BaseModel):
    username: str = Field(
        ...,
        json_schema_extra="john"
    )
    email: str = Field(
        ...,
        json_schema_extra="john@example.com"
    )


class Token(BaseModel):
    access_token: str = Field(
        ...,
        json_schema_extra="fake-token"
    )
    token_type: str = Field(
        ...,
        json_schema_extra="bearer"
    )


class SecureData(BaseModel):
    data: str = Field(
        ...,
        json_schema_extra="This is secure data"
    )


# Dependencies
def get_current_user(
    token: str = Depends(oauth2_scheme)
) -> User:
    if token == "fake-token":
        return User(
            username="john",
            email="john@example.com"
        )
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Invalid token",
        response_model=ErrorResponse,
    )


def get_api_key(
    api_key: str = Security(api_key_scheme)
) -> str:
    if api_key == "fake-api-key":
        return api_key
    raise HTTPException(
        status_code=status.HTTP_403_FORBIDDEN,
        detail="Invalid API Key",
        response_model=ErrorResponse,
    )


@pytest.fixture
def app():
    app = FastAPI()

    @app.get(
        path="/items/{item_id}",
        summary="Get an item",
        response_model=ResponseItem,
        status_code=status.HTTP_200_OK,
        tags=["Items"],
        responses={
            status.HTTP_200_OK: {
                "description": "Successful Response",
                "model": ResponseItem,
            },
            status.HTTP_404_NOT_FOUND: {
                "description": "Item not found",
                "model": ErrorResponse,
            },
        },
    )
    async def read_item(
        item_id: int = Path(
            ...,
            description="The ID of the item to retrieve"
        ),
        q: Optional[str] = Query(
            None,
            description="Query string for the item"
        ),
        user_agent: Optional[str] = Header(
            None,
            description="User-Agent header"
        ),
        cookie_id: Optional[str] = Cookie(
            None,
            description="Cookie ID"
        ),
    ) -> ResponseItem:
        if item_id == 1:
            return ResponseItem(
                id=item_id,
                name="Item name",
                description="Item description"
            )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Item not found",
            response_model=ErrorResponse,
        )

    @app.post(
        path="/items/",
        summary="Create an item",
        response_model=ResponseItem,
        status_code=status.HTTP_201_CREATED,
        tags=["Items"],
        responses={
            status.HTTP_201_CREATED: {
                "description": "Successful Response",
                "model": ResponseItem,
            },
            status.HTTP_401_UNAUTHORIZED: {
                "description": "Unauthorized",
                "model": ErrorResponse,
            },
        },
    )
    async def create_item(
        item: Item = Body(
            ...,
            examples={
                "default": {
                    "summary": "An example item",
                    "value": {
                        "name": "Example item",
                        "description": "Example description",
                    },
                }
            },
        ),
        authorization: Optional[str] = Header(
            None,
            description="Authorization token"
        ),
    ) -> ResponseItem:
        if authorization != "Bearer test-token":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Unauthorized",
                response_model=ErrorResponse,
            )
        return ResponseItem(
            id=1, **item.model_dump()
        )

    @app.post(
        path="/token",
        summary="Get a token",
        response_model=Token,
        status_code=status.HTTP_200_OK,
        tags=["Auth"],
        responses={
            status.HTTP_200_OK: {
                "description": "Successful Response",
                "model": Token,
            },
            status.HTTP_400_BAD_REQUEST: {
                "description": "Invalid credentials",
                "model": ErrorResponse,
            },
        },
    )
    async def login(
        form_data: OAuth2PasswordRequestForm = Depends()
    ) -> Token:
        if form_data.username == "john" and form_data.password == "secret":
            return Token(
                access_token="fake-token",
                token_type="bearer",
            )
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Invalid credentials",
            response_model=ErrorResponse,
        )

    @app.get(
        path="/users/me",
        summary="Get current user",
        response_model=User,
        status_code=status.HTTP_200_OK,
        tags=["Users"],
        dependencies=[
            Depends(
                get_current_user
            )
        ],
        responses={
            status.HTTP_200_OK: {
                "description": "Successful Response",
                "model": User,
            },
            status.HTTP_401_UNAUTHORIZED: {
                "description": "Unauthorized",
                "model": ErrorResponse,
            },
        },
    )
    async def read_users_me(
        current_user: User = Depends(
            get_current_user
        )
    ) -> User:
        return current_user

    @app.get(
        path="/secure-data",
        summary="Get secure data",
        response_model=SecureData,
        status_code=status.HTTP_200_OK,
        tags=["Secure"],
        dependencies=[
            Security(
                get_api_key
            )
        ],
        responses={
            status.HTTP_200_OK: {
                "description": "Successful Response",
                "model": SecureData,
            },
            status.HTTP_403_FORBIDDEN: {
                "description": "Forbidden",
                "model": ErrorResponse,
            },
        },
    )
    async def get_secure_data(
        api_key: str = Security(
            get_api_key
        )
    ) -> SecureData:
        return SecureData(
            data="This is secure data"
        )

    return app


@pytest.fixture
def client(app):
    return TestClient(app)


def test_generate_postman_collection(
    app,
    client
):
    output_file = "postman_collection.json"
    generate_postman_collection(
        app,
        str(output_file),
        "Test API",
        "http://testserver"
    )

    with open(output_file) as f:
        collection = json.load(f)

    schema = (
        "https://schema.getpostman.com/json/collection/v2.1.0/"
        "collection.json"
    )

    assert collection["info"]["name"] == "Test API"
    assert collection["info"]["schema"] == schema
    assert "item" in collection
    assert len(collection["item"]) > 0

    folder_names = [
        folder["name"]
        for folder in collection["item"]
    ]
    assert "Items" in folder_names
    assert "Auth" in folder_names
    assert "Users" in folder_names
    assert "Secure" in folder_names

    items_folder = next(
        folder
        for folder in collection["item"]
        if folder["name"] == "Items"
    )
    assert len(items_folder["item"]) == 2

    url = "http://testserver/items/{item_id}"

    get_item = items_folder["item"][0]
    assert get_item["name"] == "read_item"
    assert get_item["request"]["url"] == url
    assert get_item["request"]["method"] == "GET"
    assert get_item["request"]["description"] == "Get an item"
    assert get_item["request"]["header"] == [
        {
            "key": "user_agent",
            "value": "{{user_agent}}",
        }
    ]
    assert get_item["request"]["body"]["mode"] == "raw"
    assert get_item["request"]["body"]["raw"] == {}
    assert sorted(
        get_item["request"]["params"],
        key=lambda x: x["name"]
    ) == sorted(
        [
            {
                "name": "item_id",
                "in": "path",
                "required": True,
                "schema": {
                    "type": "int",
                    "description": "The ID of the item to retrieve",
                    "default": None,
                    "example": "",
                },
            },
            {
                "name": "q",
                "in": "query",
                "required": False,
                "schema": {
                    "type": "Optional",
                    "description": "Query string for the item",
                    "default": "",
                    "example": "",
                },
            },
        ],
        key=lambda x: x["name"],
    )
    assert get_item["request"]["responses"] == {
        "200": {
            "description": "Successful Response",
            "content": {
                "application/json": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            "description": {
                                "anyOf": [
                                    {
                                        "type": "string",
                                    },
                                    {
                                        "type": "null",
                                    },
                                ],
                                "default": None,
                                "title": "Description",
                            },
                            "id": {
                                "title": "Id",
                                "type": "integer"
                            },
                            "name": {
                                "title": "Name",
                                "type": "string",
                            },
                        },
                        "required": [
                            "name",
                            "id"
                        ],
                        "title": "ResponseItem",
                    }
                }
            },
        },
        "404": {
            "description": "Item not found",
            "content": {
                "application/json": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            "detail": {
                                "title": "Detail",
                                "type": "string",
                            }
                        },
                        "required": ["detail"],
                        "title": "ErrorResponse",
                    }
                }
            },
        },
    }

    url = "http://testserver/items/"

    create_item = items_folder["item"][1]
    assert create_item["name"] == "create_item"
    assert create_item["request"]["url"] == url
    assert create_item["request"]["method"] == "POST"
    assert create_item["request"]["description"] == "Create an item"
    assert create_item["request"]["header"] == [
        {
            "key": "authorization",
            "value": "{{authorization}}",
        }
    ]
    assert create_item["request"]["body"]["mode"] == "raw"
    assert create_item["request"]["body"]["raw"] == {
        "name": "Example item",
        "description": "Example description",
    }
    assert create_item["request"]["params"] == []
    assert create_item["request"]["responses"] == {
        "201": {
            "description": "Successful Response",
            "content": {
                "application/json": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            "description": {
                                "anyOf": [
                                    {
                                        "type": "string",
                                    },
                                    {
                                        "type": "null",
                                    },
                                ],
                                "default": None,
                                "title": "Description",
                            },
                            "id": {
                                "title": "Id",
                                "type": "integer",
                            },
                            "name": {
                                "title": "Name",
                                "type": "string",
                            },
                        },
                        "required": [
                            "name",
                            "id"
                        ],
                        "title": "ResponseItem",
                    }
                }
            },
        },
        "401": {
            "description": "Unauthorized",
            "content": {
                "application/json": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            "detail": {
                                "title": "Detail",
                                "type": "string",
                            }
                        },
                        "required": ["detail"],
                        "title": "ErrorResponse",
                    }
                }
            },
        },
    }

    response = client.get(
        "/items/1",
        headers={
            "user_agent": "test-agent"
        },
        params={
            "q": "test-query"
        },
    )
    assert response.status_code == 200
    assert response.json() == {
        "id": 1,
        "name": "Item name",
        "description": "Item description",
    }

    response = client.post(
        "/items/",
        json={
            "name": "test",
            "description": "test description"
        },
        headers={
            "authorization": "Bearer test-token",
        },
    )
    assert response.status_code == 201
    assert response.json() == {
        "id": 1,
        "name": "test",
        "description": "test description",
    }


def test_generate_postman_collection_with_auth(
    app,
    client
):
    output_file = "postman_collection_with_auth.json"
    generate_postman_collection(
        app,
        str(output_file),
        "Test API with Auth",
        "http://testserver"
    )

    with open(output_file) as f:
        collection = json.load(f)

    assert collection["auth"]["type"] == "bearer"
    assert collection["auth"]["bearer"][0]["key"] == "token"
    assert collection["auth"]["bearer"][0]["value"] == "{{access_token}}"
    assert collection["auth"]["bearer"][0]["type"] == "string"

    folder_names = [
        folder["name"]
        for folder in collection["item"]
    ]
    assert "Items" in folder_names
    assert "Auth" in folder_names
    assert "Users" in folder_names
    assert "Secure" in folder_names

    items_folder = next(
        folder
        for folder in collection["item"]
        if folder["name"] == "Items"
    )
    assert len(items_folder["item"]) == 2

    get_item = items_folder["item"][0]
    assert sorted(
        get_item["request"]["params"],
        key=lambda x: x["name"],
    ) == sorted(
        [
            {
                "name": "item_id",
                "in": "path",
                "required": True,
                "schema": {
                    "type": "int",
                    "description": "The ID of the item to retrieve",
                    "default": None,
                    "example": "",
                },
            },
            {
                "name": "q",
                "in": "query",
                "required": False,
                "schema": {
                    "type": "Optional",
                    "description": "Query string for the item",
                    "default": "",
                    "example": "",
                },
            },
        ],
        key=lambda x: x["name"],
    )

    response = client.get(
        "/items/1",
        headers={
            "Authorization": "Bearer test-token"
        },
    )
    assert response.status_code == 200
    assert response.json() == {
        "id": 1,
        "name": "Item name",
        "description": "Item description",
    }

    response = client.post(
        "/items/",
        json={
            "name": "test",
            "description": "test description",
        },
        headers={
            "Authorization": "Bearer test-token",
        },
    )
    assert response.status_code == 201
    assert response.json() == {
        "id": 1,
        "name": "test",
        "description": "test description",
    }



from setuptools import setup, find_packages
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="fast-man",
    version="0.1.7",
    packages=find_packages(),
    install_requires=[
        "fastapi",
        "pydantic",
        "requests"
    ],
    entry_points={
        "console_scripts": [
            "fast-man=fast_man.converter:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    long_description=README,
    long_description_content_type="text/markdown",
)
